#include "BinaryTree.h"
#include <iso646.h>
#include <stdlib.h>

enum BinaryTreeNodePosition {
	BINARYTREENODEPOSITION_SELF,
	BINARYTREENODEPOSITION_LEFT,
	BINARYTREENODEPOSITION_RIGHT,
};

typedef struct BinaryTreeWalker_s {
	BinaryTreeNode* node;
	enum BinaryTreeNodePosition position;
} BinaryTreeWalker;

static size_t _BinaryTree_TreeNodeSize(BinaryTree* tree)
{
	return sizeof(BinaryTreeNode) + tree->value_size;
}

static inline size_t _BinaryTree_DepthOf(BinaryTreeNode* node)
{
	return node == NULL ? 0 : node->depth;
}

static BinaryTreeNode* _BinaryTree_CreateNode(BinaryTree* tree, void* value)
{
	BinaryTreeNode* node = Allocator_Allocate(tree->allocator, _BinaryTree_TreeNodeSize(tree));
	if (node != NULL) {
		memcpy(node->value, value, tree->value_size);
		node->depth = 1;
		node->left = NULL;
		node-> right = NULL;
	}

	return node;
}

int BinaryTree_Create(BinaryTree* target, BinaryTreeComparator compare, void* xdata, size_t value_size, allocator_t* allocator)
{
	target->root = NULL;
	target->compare = compare;
	target->xdata = xdata;
	target->allocator = allocator;
	target->value_size = value_size;

	int array_code = DynamicArray_Create(&target->walker_stack, sizeof(BinaryTreeWalker), 8, allocator);
	if (array_code) return array_code;

	return EXIT_SUCCESS;
}

static void _BinaryTreeNode_UpdateDepth(BinaryTreeNode* node)
{
	size_t depth_left = 0, depth_right = 0;

	if (node->left != NULL) {
		depth_left = node->left->depth;
	}
	if (node->right != NULL) {
		depth_right = node->right->depth;
	}

	if (depth_left > depth_right) {
		node->depth = depth_left;
	} else {
		node->depth = depth_right;
	}

	node->depth += 1;
}

static void _BinaryTree_UpdateDepths(DynamicArray* parents)
{
	for (size_t i = parents->reserved - 1; i != SIZE_MAX; i--) {
		BinaryTreeWalker* walker = DynamicArray_GetPointer(parents, i);
		_BinaryTreeNode_UpdateDepth(walker->node);
	}
}

static int _BinaryTree_ValueWalk(BinaryTree* tree, void* value)
{
	BinaryTreeNode** parent = &tree->root;
	tree->walker_stack.reserved = 0;

	while ((*parent) != NULL) {
		BinaryTreeWalker* walker;
		DynamicArray_AppendEmpty(&tree->walker_stack, (void**) &walker);
		// Can't be NULL because the size was just increased

		walker->node = *parent;

		int comparison = tree->compare(value, (*parent)->value, tree->xdata);

		if (comparison == 0) {
			// equal
			walker->position = BINARYTREENODEPOSITION_SELF;
			return EXIT_SUCCESS;
		} else if (comparison < 0) {
			parent = &(*parent)->left;
			walker->position = BINARYTREENODEPOSITION_LEFT;
		} else {
			parent = &(*parent)->right;
			walker->position = BINARYTREENODEPOSITION_RIGHT;
		}
	}

	return ENOTFOUND;
}

static int _BinaryTree_Insert(BinaryTree* tree, BinaryTreeNode* node)
{
	int walk_code = _BinaryTree_ValueWalk(tree, node->value);

	if (walk_code != ENOTFOUND) {
		return EEXIST;
	}

	BinaryTreeWalker* bottom_walker = DynamicArray_GetPointer(&tree->walker_stack, tree->walker_stack.reserved - 1);
	if (bottom_walker == NULL) {
		// root node is empty
		tree->root = node;
		return EXIT_SUCCESS;
	} else {
		switch (bottom_walker->position) {
			case BINARYTREENODEPOSITION_SELF:
				// unreachable code
				return EXIT_FAILURE;

			case BINARYTREENODEPOSITION_LEFT:
				bottom_walker->node->left = node;
				break;

			case BINARYTREENODEPOSITION_RIGHT:
				bottom_walker->node->right = node;
				break;
		}
	}

        _BinaryTree_UpdateDepths(&tree->walker_stack);
	_BinaryTreeNode_UpdateDepth(node);

	return EXIT_SUCCESS;
}

static int _BinaryTree_DestroyNode(BinaryTree* tree, BinaryTreeNode* node)
{
	Allocator_Free(tree->allocator, node, sizeof(*node) + tree->value_size);

	return EXIT_SUCCESS;
}

static BinaryTreeNode* _BinaryTreeNode_SmallestNodeParent(BinaryTreeNode* node)
{
	if (node->left == NULL) {
		return NULL;
	}
	
	BinaryTreeNode* parent = node;
	while (parent->left != NULL && (parent->left->left != NULL)) {
		parent = parent->left;
	}

	return parent;
}

static int _BinaryTreeNode_UnlinkNode(BinaryTreeNode** nodep)
{
	size_t node_depth = _BinaryTree_DepthOf(*nodep);
	if (node_depth == 1) {
		// leaf node without children
		*nodep = NULL;
	} else {
		if ((*nodep)->left == NULL) {
			// only right node active
			*nodep = (*nodep)->right;
		} else if ((*nodep)->right == NULL) {
			// only left node active
			*nodep = (*nodep)->left;
		} else {
			// both right and left node active
			BinaryTreeNode* smallest_parent = _BinaryTreeNode_SmallestNodeParent(*nodep);
			BinaryTreeNode* smallest_node   = smallest_parent->left;

			smallest_parent->left = smallest_node->right;
			smallest_node->right  = (*nodep)->right;
			smallest_node->left   = (*nodep)->left;
			*nodep = smallest_node;

			_BinaryTreeNode_UpdateDepth(smallest_parent);
		}
	}

	return EXIT_SUCCESS;
}

static int _BinaryTree_Remove(BinaryTree* self, void* value)
{
	int walk_code = _BinaryTree_ValueWalk(self, value);

	if (walk_code == ENOTFOUND) {
		return ENOTFOUND;
	}

	BinaryTreeWalker* bottom_walker = DynamicArray_GetPointer(&self->walker_stack, self->walker_stack.reserved - 2);

	BinaryTreeNode* preserved_pointer;
	if (bottom_walker == NULL) {
		// root node
		preserved_pointer = self->root;
		_BinaryTreeNode_UnlinkNode(&self->root);
		self->walker_stack.reserved = 0;
	} else {
		BinaryTreeNode* parent_node = bottom_walker->node;
		switch (bottom_walker->position) {
			case BINARYTREENODEPOSITION_SELF:
				return EXIT_FAILURE;

			case BINARYTREENODEPOSITION_LEFT:
				preserved_pointer = parent_node->left;
				_BinaryTreeNode_UnlinkNode(&parent_node->left);
				break;

			case BINARYTREENODEPOSITION_RIGHT:
				preserved_pointer = parent_node->right;
				_BinaryTreeNode_UnlinkNode(&parent_node->right);
				break;
                }

		bottom_walker->position = BINARYTREENODEPOSITION_SELF;
        }
	DynamicArray_Remove(&self->walker_stack, self->walker_stack.reserved - 1);

	_BinaryTree_DestroyNode(self, preserved_pointer);

	return EXIT_SUCCESS;
}

int _BinaryTree_GrowWalkerStack(BinaryTree* tree)
{
	size_t depth = _BinaryTree_DepthOf(tree->root);
	if (tree->walker_stack.capacity < depth + 1) {
		 return DynamicArray_Resize(&tree->walker_stack, tree->walker_stack.capacity * 2);
	}

	return EXIT_SUCCESS;
}

// right is heavier, shift it to the left
static void _BinaryTreeNode_RotateLeft(BinaryTreeNode** node)
{
	BinaryTreeNode* right = (*node)->right;

	(*node)->right = right->left;
	right->left    = *node;
	*node          = right;
}

// left is heavier, shift it to the right
static void _BinaryTreeNode_RotateRight(BinaryTreeNode** node)
{
	BinaryTreeNode* left = (*node)->left;

	(*node)->left = left->right;
	left->right   = *node;
	*node         = left;
}

static void _BinaryTreeNode_Rebalance(BinaryTreeNode** node)
{
	size_t left_depth  = _BinaryTree_DepthOf(node[0]->left);
	size_t right_depth = _BinaryTree_DepthOf(node[0]->right);

	if (left_depth < right_depth && right_depth - left_depth > 1) {
		_BinaryTreeNode_RotateLeft(node);
		_BinaryTreeNode_UpdateDepth((*node)->left); // old node
		_BinaryTreeNode_UpdateDepth(*node); // rotated node
	} else if (right_depth < left_depth && left_depth - right_depth > 1) {
		_BinaryTreeNode_RotateRight(node);
		_BinaryTreeNode_UpdateDepth((*node)->right); // old node
		_BinaryTreeNode_UpdateDepth(*node); // rotated node
	}
}

static void _BinaryTree_RebalanceRoot(BinaryTree* tree)
{
	_BinaryTreeNode_Rebalance(&tree->root);
}

static void _BinaryTree_Rebalance(BinaryTree* tree)
{
	for (size_t i = tree->walker_stack.reserved - 1; i + 1 != 0; i--) {
		BinaryTreeWalker* walker = DynamicArray_GetPointer(
			&tree->walker_stack,
			i
		);
		BinaryTreeNode** nodep = NULL;
		switch (walker->position) {
			case BINARYTREENODEPOSITION_LEFT:
				nodep = &walker->node->left;
				break;

			case BINARYTREENODEPOSITION_RIGHT:
				nodep = &walker->node->right;
				break;

			case BINARYTREENODEPOSITION_SELF:
				if (i == 0) {
					nodep = &tree->root;
				} else {
					BinaryTreeWalker* parent_walker = DynamicArray_GetPointer(
						&tree->walker_stack,
						i - 1
					);
					if (parent_walker->position == BINARYTREENODEPOSITION_LEFT) {
						nodep = &parent_walker->node->left;
					} else {
						// right, self is not allowed
						nodep = &parent_walker->node->right;
					}
				}
				break;

		}

		_BinaryTreeNode_Rebalance(nodep);
	}


	_BinaryTree_RebalanceRoot(tree);
}

int BinaryTree_Insert(BinaryTree* tree, void* value)
{
	if (tree == NULL) return EDESTADDRREQ;
	if (value == NULL) return EINVAL;

	if (_BinaryTree_GrowWalkerStack(tree)) {
		return ENOMEM;
	}

	BinaryTreeNode* new_node = _BinaryTree_CreateNode(tree, value);
	if (new_node == NULL) return ENOMEM;

	if (_BinaryTree_Insert(tree, new_node)) {
		_BinaryTree_DestroyNode(tree, new_node);
		return EXIT_FAILURE;
	}
	_BinaryTree_Rebalance(tree);

	return EXIT_SUCCESS;
}

int BinaryTree_Remove(BinaryTree* self, void* value)
{
	if (self == NULL) return EDESTADDRREQ;
	if (value == NULL) return EINVAL;

	int remove_code = _BinaryTree_Remove(self, value);
	if (remove_code) {
		return remove_code;
	}

	_BinaryTree_Rebalance(self);

	return EXIT_SUCCESS;
}

int BinaryTree_ForEachPostOrder(BinaryTree* tree, BinaryTreeNodeFunction action, void* context)
{
	tree->walker_stack.reserved = 0;

	{
		BinaryTreeWalker* root_walker;
		DynamicArray_AppendEmpty(&tree->walker_stack, (void**) &root_walker);

		root_walker->node = tree->root;
		root_walker->position = BINARYTREENODEPOSITION_LEFT;
	}

	while (tree->walker_stack.reserved != 0) {
		BinaryTreeWalker* current = DynamicArray_GetPointer(&tree->walker_stack, tree->walker_stack.reserved - 1);
		if (current->node == NULL) {
			tree->walker_stack.reserved--; // remove last item
			continue;
		}

		BinaryTreeNode* append = NULL;

		switch (current->position) {

			case BINARYTREENODEPOSITION_SELF:
			{
				int error = action(context, current->node);
				if (error) return ECANCELED;

				tree->walker_stack.reserved--;
				break;
			}

			case BINARYTREENODEPOSITION_LEFT:
				current->position = BINARYTREENODEPOSITION_RIGHT;
				append = current->node->left;
				break;

			case BINARYTREENODEPOSITION_RIGHT:
				current->position = BINARYTREENODEPOSITION_SELF;
				append = current->node->right;
				break;

		}

		if (append != NULL) {
			BinaryTreeWalker* new_walker;
			DynamicArray_AppendEmpty(&tree->walker_stack, (void**) &new_walker);
			new_walker->node = append;
			new_walker->position = BINARYTREENODEPOSITION_LEFT;
		}
	}

	return EXIT_SUCCESS;
}

void BinaryTree_Destroy(BinaryTree* tree)
{
	BinaryTree_ForEachPostOrder(tree, (BinaryTreeNodeFunction) _BinaryTree_DestroyNode, tree);
	DynamicArray_Destroy(&tree->walker_stack);
	return;
}
